## Copyright (C) 2012   Robert T. Short     <rtshort@ieee.org>
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; If not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn {Function File} {[@var{b},@var{ec},@var{err}]} = imarcumQ(@var{Q},@var{a},@var{M},@var{tol}=1e-10,@var{maxiter}=100,@var{initguess}=[])
##
## @noindent
## Compute the inverse of Marcum's Q function given the
## noncentrality parameter.
##
## @noindent
## If the input arguments are commensurate vectors, this function
## will produce a table of values (see tablify).
##
## @noindent
## This is performed using the Newton-Raphson method on the
## log of the error probabilities.
##
## Parameter descriptions 
## @table @var
## @item Q
## The value of the Q function. This may be a vector. 
## @item a
## Noncentrality parameter.
## @item M
## The order of the Q function.
## @item tol
## The convergence tolerance.
## @item maxiter
## Maximum iterations before quitting.
## @item initguess
## An initial estimate.  This may be a scalar or a vector the
## same size as Q.  A sensible default will be used if this
## parameter is not present.
##
## @item b
## The inverse.
## @item ec
## Zero if convergence was successful, 1 if not.
## @item err
## Vector containing convergence information.  If ec
## is nonzero, this will be greater than tol.
## @end table
##
## @noindent
## Notes.
##
## @noindent
## This function is robust over a fairly broad range of values, but will
## fail to converge on the extreme tails.  This is because the
## derivative is nearly zero for values of the Q function less than
## 1e-12 or so.  Note that the result will normally be a correct result
## anyway.
##
## @noindent
## This routine has been tested for M=50, but the Marcum Q routine
## itself is questionable even at that level.
##
## @noindent
## The tolerance and convergence parameters, tol, maxiter, and initguess
## should only be changed if you have very good knowledge of the
## problem. The defaults are very solid.
##
## @noindent
## Reference: Marcum, "Tables of Q Functions", Rand Corporation.
##
## @noindent
## Reference: R.T. Short, "Computation of Noncentral Chi-squared
## and Rice Random Variables", www.phaselockedsystems.com/publications
##
## @seealso{marcumQ,tablify}
##
## @end deftypefn
  
## Author:  Robert T. Short (rtshort@ieee.org)
## Description:  Inverse of Marcum's Q function.
function [b,ec,err,cnt]=imarcumQ(mQ,a=0,M=1,tol=1e-10,maxiter=50,initguess=[])

  if(nargin<2)
    print_usage();
  end
  if( (any(mQ<0)) || (any(mQ)>1) )
    error('imarcumQ: probabilities must be between 0 and 1');
  end

  mqtol = tol/100;
  [mQ,a,M] = tablify(mQ,a,M);
  if ( isempty(initguess) )
    initguess = arrayfun(@bInit,a,M);
  end

  [b,ec,err,cnt] = arrayfun(@imq,mQ,a,M,tol,mqtol,maxiter,initguess);

end

% Function to perform the actual iteration.
function [b,ec,err,cnt]=imq(mQ,a,M,tol,mqtol,maxiter,initguess)
  b = initguess;
  cnt = 0;
  do
    cnt++;
    g  = marcumQ(a,b,M,mqtol);
    gp = fprime(a,b,M);
    if (gp>=0)
      delta = NaN;
      done  = true;
    else
      delta = g*(log(g)-log(mQ))/gp;
      b     = b - delta;
      done  = abs(delta) < tol;
    end
  until ( done || (cnt>maxiter) || (gp>=0) )
  ec = ~done;
  err = delta;

end

% Use the mean as the starting point.
function init=bInit(a,M)
  t = 1;
  S = 1;
  k = 0;
  f = a^2/2;
  do
    k++;
    t = t*(M+k-1/2)*f/(k*(M+k-1));
    S = S+t;
  until(t<1e-4)
  init = sqrt(2)*e^(-f)*gamma(M+1/2)/gamma(M) * S;
end
% derivative of the Q function w.r.t. a.
function fp = fprime(a,x,M)
  if(a==0)
    fp = -(x.^(2*M-1)) .* exp(-x.^2/2) / (2^(M-1)*gamma(M));
  else
    fp = -x.*exp(-(a-x).^2/2).*(x./a).^(M-1).*besseli(M-1,a.*x,1);
  end
end


% Some tests of situations that are normal and well-behaved.
% These tests are based on Marcum's Table of Q Functions.

%!test
%! b = 0.1;
%! a = [0.0; 0.1; 1.0; 2.0; 3.0; 4.0; 5.0];
%! Q = [0.9950124791927; 0.9950186941452; 0.9969711359363;
%!      0.9993216347111; 0.9999439685077; 0.999998293215; 
%!      0.9999999808268                                   ];
%! [bhat,ec] = imarcumQ(Q,a);
%! assert(marcumQ(a,bhat),Q,1e-10)
%! assert(~any(ec))

%!test
%! b = 1.1;
%! a = [0.0; 0.1; 1.0; 2.0; 3.0; 4.0; 5.0; 6.0; 7.0];
%! Q = [0.5460744266397; 0.5464872154453; 0.6853770960717;
%!      0.8980728412671; 0.985456591056;  0.9991360618694; 
%!      0.9999794670311; 0.9999998096743; 0.999999999323  ];
%! [bhat,ec] = imarcumQ(Q,a);
%! assert(marcumQ(a,bhat),Q,1e-10)
%! assert(~any(ec))

%!test
%! b = 6.1;
%! a = [0.0; 0.1; 1.0; 2.0; 3.0; 4.0; 5.0; 6.0; 7.0; 8.0; 9.0; 10.0; 11.0; 12.0];
%! Q = [8.316702456828e-09; 8.511123433478e-09; 4.350607714669e-07; 
%!      3.710053912084e-05; 0.00141987240829;   0.02281152287179; 
%!      0.1564576914076;    0.4932293738009;    0.8356569041839; 
%!      0.9756729382385;    0.9984977269359;    0.9999630791264; 
%!      0.9999996481211;    0.9999999987192                        ];
%! [bhat,ec] = imarcumQ(Q,a);
%! assert(marcumQ(a,bhat),Q,1e-10)
%! assert(~any(ec))

%!test
%! b = 12.1;
%! a = [6.0; 7.0; 8.0; 9.0; 10.0; 11.0; 12.0; 13.0; 14.0; 15.0; 16.0; 17.0; 18.0];
%! Q = [7.59262905957e-10; 2.252881128997e-07; 2.566812843627e-05; 
%!      0.001135888814819; 0.01996110072598;   0.1453417510951; 
%!      0.4766918167483;   0.8263656930001;    0.9737140366318; 
%!      0.9983433918087;   0.9999585486772;    0.999999598679; 
%!      0.999999998519                                            ];
%! [bhat,ec] = imarcumQ(Q,a);
%! assert(marcumQ(a,bhat),Q,1e-10)
%! assert(~any(ec))

%!test
%! b = 19.1;
%! a = [13.0; 14.0; 15.0; 16.0; 17.0; 18.0; 19.0; 20.0; 21.0; 22.0; 23.0; 24.0];
%! Q = [6.457774855813e-10; 1.994019887587e-07; 2.345462888617e-05; 
%!      0.001065215148586;  0.01912088854856;   0.1416299660763; 
%!      0.4706081749148;    0.8226707673073;    0.9728832469933;
%!      0.9982742912635;    0.9999564239705;    0.9999995745404    ];
%! [bhat,ec] = imarcumQ(Q,a);
%! assert(marcumQ(a,bhat),Q,1e-10)
%! assert(~any(ec))

%!test
%! M = 10;
%! b = 1.1;
%! a = [0.0; 0.1; 1.0; 2.0];
%! Q = [0.9999999989541; 0.9999999989553; 0.999999999348; 0.9999999998422];
%! bhat = imarcumQ(Q,a,M);
%! assert(marcumQ(a,bhat,M),Q,1e-10)

%!test
%! M = 10;
%! b = 6.1;
%! a = [0.0; 0.1; 1.0; 2.0; 3.0; 4.0; 5.0; 6.0; 7.0; 8.0; 9.0; 10.0; 11.0];
%! Q = [0.01104157070746; 0.01105581308631; 0.01777580345003; 
%!      0.05274886429141; 0.1683856008584;  0.4196427703565; 
%!      0.7308256668127;  0.9295849276456;  0.990793135024; 
%!      0.9994453786769;  0.9999854218203;  0.9999998389046; 
%!      0.999999999271];
%! bhat = imarcumQ(Q,a,M);
%! assert(marcumQ(a,bhat,M),Q,1e-10)

%!test
%! M = 10;
%! b = 12.1;
%! a = [5.0; 6.0; 7.0; 8.0; 9.0; 10.0; 11.0; 12.0; 13.0; 14.0; 15.0; 16.0; 17.0];
%! Q = [1.589925776109e-09; 2.72227733671e-07; 2.235995205935e-05; 
%!      0.000840059959402;  0.01415962506148;  0.107998918286; 
%!      0.3911808587271;    0.7546307449195;   0.9513536752891; 
%!      0.995732411008;     0.9998437423414;   0.9999976912448; 
%!      0.9999999865193];
%! bhat = imarcumQ(Q,a,M);
%! assert(marcumQ(a,bhat,M),Q,1e-10)

%!test
%! M = 10;
%! b = 19.1;
%! a = [13.0; 14.0; 15.0; 16.0; 17.0; 18.0; 19.0; 20.0; 21.0; 22.0; 23.0; 24.0];
%! Q = [1.890176912413e-08; 3.072919819294e-06; 0.0002000684969216; 
%!      0.00527705532195;   0.05786956598269;   0.2784262033404; 
%!      0.6550169398204;    0.9171635452135;    0.9912112775139;  
%!      0.9996149060474;    0.9999932808234;    0.999999954287];
%! bhat = imarcumQ(Q,a,M);
%! assert(marcumQ(a,bhat,M),Q,1e-10)

%  A couple of examples of places where convergence apparently
%  fails (but really doesn't).  These are good examples.

%!test
%! a=5; b=0; M=5; Q=marcumQ(a,b,M);
%! [bhat,ec] = imarcumQ(Q,a,M);
%
%  The algorithm converged.
%! assert(ec==0);
%
%  The resulting value of bhat is not what we started with.
%! assert(b!=bhat);
%
%  But the value of the Q function at the converged
%  value is the same as the original value of the 
%  Q function (and therefore the algorithm did
%  converge properly).  The solution is not unique,
%  at least to within machine precision.
%! assert(marcumQ(a,bhat,M),Q,1e-10);

%!test
%! a=5; b=14; M=5; Q=marcumQ(a,b,M);
%! [bhat,ec] = imarcumQ(Q,a,M);
%
%  The algorithm may or may not converge but
%  the resulting value of b is not even close.
%! assert(bhat!=b);
%
%  However, the value of the Q function at the
%  resulting value of bhat is (within tolerance)
%  the same as the value of the Q function at
%  the original value.  Therefore, the answer is
%  in fact correct and the solution is not unique
%  to within the tolerance of the Marcum Q function
%  itself.
%! assert(marcumQ(a,bhat,M),Q,1e-10);
